function Figure10B
%% Generates the stochastic convergence plot in figure 10B.

% I used export_fig to export my figures. Setting export_fig = 0 will disable
% saving of the figures with export_fig. You can download export_fig from the 
% MathWorks file exchange.

%% Setup
user= 'USER';
project_path = ['/home/' user '/chaste-source/StochasticPhenomenologicalModels'];
output_path = [ '/tmp/' user '/testoutput' ];
matlab_path = [ project_path '/Matlab' ];
% export_fig?
exportFig = 0;

% variables under investigation
conductivityScales = [1.0 0.05];
printTimeSteps = [1.0 0.1 0.05];
refinementLevels  = [0.01 0.005 0.0025 0.00125 0.000625];

% mesh properties
spatialDiscretisation = 0.01;
sideLength = 0.2;
numElemX = sideLength/spatialDiscretisation;
numElemY = sideLength/spatialDiscretisation;
numNodesX = numElemX+1;
numNodesY = numElemY+1;

% number of nodes
numNodes = numNodesX*numNodesY;
% array to hold results
resultsArray = zeros(length(conductivityScales), length(printTimeSteps), length(refinementLevels)-1);

%% Load Data
% change to output directory
cd( [ output_path '/StochasticConvergenceTest' ] )

% loop over conductivity scales
for conductivityIdx = 1:length(conductivityScales)
    % loop over printing time steps (does not appear in figure in paper as made no difference)
    for printTimeStepIdx = 1:length(printTimeSteps)
        % load the reference APD file at the smallest time step
        referenceApdFile = ['RefinementLevel_' num2str(length(refinementLevels)) '_PrintTimeStep_' num2str(printTimeSteps(printTimeStepIdx))...
                                '_CondScale_' num2str(conductivityScales(conductivityIdx)) '/output/Apd_90_0pt40_Map.dat'];
        referenceData = dlmread(referenceApdFile, '', [0 0 numNodes-1 0]);
        referenceApdMap = reshape(referenceData, numNodesY, numNodesX); 
        % load data for the differnt refinement levels (stochastic time steps)        
        for refinementLevelIdx = 1:length(refinementLevels)-1
            % load file for current refinement level
            currentApdFile = ['RefinementLevel_' num2str(refinementLevelIdx) '_PrintTimeStep_' num2str(printTimeSteps(printTimeStepIdx))...
                                '_CondScale_' num2str(conductivityScales(conductivityIdx)) '/output/Apd_90_0pt40_Map.dat']; 
            currentData = dlmread(currentApdFile, '', [0 0 numNodes-1 0]);
            currentApdMap = reshape(currentData, numNodesY, numNodesX); 
            % calculate absolute difference between current value and reference value
            absDifference = abs(currentApdMap-referenceApdMap);
			% find maximum difference
            maxAbsDifference = max(max(absDifference));
            % load results into array
            resultsArray(conductivityIdx, printTimeStepIdx, refinementLevelIdx) = maxAbsDifference; 
            
        end % refinement levels
        
    end % print time step
    
end % conductivity

%% plot results
% change to Matlab directory
cd( matlab_path )
% calculate largest difference for axis.
globalMaxDifference = max(max(max(resultsArray)));

colourList = {'rx--','gx--','bx--'};
% open figure
fig = figure('units', 'normalized', 'outerposition', [0 0.3 1 0.7]); 
% loop over conductivities
for conductivityIdx = 1:length(conductivityScales)
    % open new subplot per conductivity
    subplot(1,length(conductivityScales), conductivityIdx)
    hold on
	% plot data per print time step n a semi log axis.
    for printTimeStepIdx = 1:length(printTimeSteps)
        colour = colourList{printTimeStepIdx};
        semilogx(refinementLevels(1:end-1), squeeze(resultsArray(conductivityIdx, printTimeStepIdx, :)), colour, 'LineWidth', 3.0, 'MarkerSize', 10);       
    end
    hold off
% cosmetics and labelling
    set(gca, 'LineWidth', 2.0)
    set(gca, 'fontsize', 20);
    ylim([0 globalMaxDifference]);
    ylabel('Abs. Difference in APD{_{90}} (ms)');
    xlabel('SDE Time step')
end % conductivites

% cosmetics
set(findall(gcf, 'Type', 'axis'), 'LineWidth', 2.0)
set(findall(gcf, 'Type', 'axis'), 'FontSize', 20)
set(findall(gcf, 'Type', 'text'), 'FontSize', 20)

if exportFig
	export_fig( 'Fig10B_StochasticConvergence.eps', '-eps', '-transparent' );
end

end
