function Figure1B
% Plot the Poincare plots in Fig 1A

% I used export_fig to export my figures. Setting export_fig = 0 will disable
% saving of the figures with export_fig. You can download export_fig from the 
% MathWorks file exchange.

%% Setup
user= 'USER';
project_path = ['/home/' user '/chaste-source/StochasticPhenomenologicalModels'];
output_path = [ '/tmp/' user '/testoutput' ];
matlab_path = [ project_path '/Matlab' ];
% export_fig?
exportFig = 0;

% Matlab script to find APD90s for single cell experimental data 
data1 = 'Data/Actionpotentials080624cell4.csv';

% Data file is labelled AP. First column is time.
AP1 = load(data1);

% Rescale transmembrane potentials
AP1(:,2:end) = 0.1*AP1(:,2:end);

% Rescale times
AP1(:,1) = 1000*AP1(:,1);

% Determine size of AP
[m1, n1] = size(AP1);

% Initialise vector for times to 90% repolarisation
T_90_1 = zeros(1,94);
% initialise vector for resting potential
rest_potential_1 = zeros(1,94);
% Set counter to 1
counter = 1;
% loop over number of action potentials
for i = 2:95
    % find maximum potential difference of AP
    APD_max = max(AP1(:,i));
    % determine time point at which this maximum occurs, k
    for j=1:m1
        if AP1(j,i) == APD_max
            k=j;
        end
    end
    % Redefine APD max as the value of the APD at time t = 0.03s, occurs at
    % point 501.
    APD_max = AP1(501,i);
    % determine time point at which 90% repolarisation is reached. Note it 
    % is necessary to rescale due to the resting potential being negative. 
    % We take the rest potential to be the mean of the first 10 values of 
    % the trace. Note that the rest potential is negative.
    rest_potential_1(counter) = mean(AP1(1:10,i));
    % loop over all time points after the maximum.  
    for j = k+1:m1
        if AP1(j,i)-rest_potential_1(counter) > 0.1*(APD_max-rest_potential_1(counter))
            
            % Time to 90% repolarisation is the difference between the time
            % at peak and the time at which the membrane potential first
            % drops below 90% of the peak.
            T_90_1(counter) = AP1(j,1)-16;
        end
    end
	% increase counter
    counter=counter+1;
end

% Cycle lengths in simulations
cycle_lengths = [1000 400 300 260 220];
% Base file names
file_bases = {'GP_Apex_1' 'GP_Apex_2' 'GP_Base_1' 'GP_Base_2'}; 

% use GP apex cell 1 for paper figure
cellNumber = 1;
   
% make figure 
figureHandle = figure;
file_base = file_bases{cellNumber};
% experimental data    
subplot(2,(length(cycle_lengths)+1)/2, 1)
plot(T_90_1(1:end-1), T_90_1(2:end), 'r+-', 'Markersize', 3.0, 'LineWidth', 1.5)
ylabel('APD_{90} n+1 (ms)')
cd(output_path)
cd('SingleCellSims/PoincarePlots')
% loop over model cycle lengths
for cl = 1:length(cycle_lengths)
	% get the deterministic and stochastic data	
	APD90s_det = load([file_base '_APD90_CL_' num2str(cycle_lengths(cl)) '.dat']);
	APD90s_stoch = load([file_base '_stoch_APD90_CL_' num2str(cycle_lengths(cl)) '.dat']);
	% Plot the simulated data    
	subplot(2,(length(cycle_lengths)+1)/2, cl+1)
    hold on
	% stochastic simulated    
	plot(APD90s_stoch(4:end-1), APD90s_stoch(5:end), 'k+-', 'Markersize', 3.0, 'LineWidth', 1.5)
	% deterministic simulated
    plot(APD90s_det(40:end-1), APD90s_det(41:end), 'bo', 'Markersize', 4.0, 'LineWidth', 2.0)   
	% label appropriately    
	if mod(cl+1,(length(cycle_lengths)+1)/2) == 1
    	ylabel('APD_{90} n+1 (ms)')
	end	
	if (cl+1 > (length(cycle_lengths)+1)/2)
    	xlabel('APD_{90} n (ms)')
    end
end
   
% Set axes properties - cosmetics 
allAxesInFigure = findall(figureHandle,'type','axes');
for ax=1:length(allAxesInFigure)
	set(allAxesInFigure(ax), 'FontSize', 16)
	set(findall(allAxesInFigure(ax), 'Type','text'), 'FontSize', 18)
	set(allAxesInFigure(ax), 'LineWidth', 1.5)
	set(allAxesInFigure(ax), 'Box', 'Off')  
end
cd(matlab_path)
% save figure
if exportFig
	export_fig( 'Fig1B_PoincarePlots.eps', '-eps', '-transparent' )
end

end

function [APD, max_V, repolarisation_threshold] = find_APD(T,Y, percentage)
% Function to find the APD from maximum upstroke velocity to 90%
% repolarisation. Used for single cell calculations in Fig 1.
% We define the APD as the time from maximum upstroke to the time of 90%
% depolarisation. Note: max upstroke is defined as point of highest
% gradient
assert(0 < percentage && percentage < 100, 'Percentage must lie between 0 and 100')
factor = (100-percentage)/100;
APD=0;
% to find the gradient
grad = diff(Y)./diff(T);
[maximum_upstroke, grad_index] = max(grad);
time_of_max_upstroke = T(grad_index+1); %+1 necessary due to nature of diff.
rest_potential = Y(1);
[max_V, max_V_index] = max(Y);
repolarisation_threshold = factor*(max_V -rest_potential) +rest_potential;
repolarisation_index_part = find(Y(max_V_index:end) < repolarisation_threshold, 1); % Finds the index in a vector of length Y(end) - max_V_index + 1
if isempty(repolarisation_index_part) % ie model doesn't reach repolarisation.
    APD = T(end)-time_of_max_upstroke;
else
    repolarisation_index = max_V_index + repolarisation_index_part-1;
    repolarisation_time = (repolarisation_threshold*(T(repolarisation_index)-T(repolarisation_index-1)) ...
        +Y(repolarisation_index)*T(repolarisation_index-1) -Y(repolarisation_index-1)*T(repolarisation_index))*(1/(Y(repolarisation_index)-Y(repolarisation_index-1)));
    APD = repolarisation_time - time_of_max_upstroke;
end

end
