function Figure2
%% Generates the statistical summary plots beat by beat for figure 2. Also 
% provides a statistical summary in a text file (Fig2Summary.dat). project_path 
% and output_path should be changed to be made appropriate to the user.

% I used export_fig to export my figures. Setting export_fig = 0 will disable
% saving of the figures with export_fig. You can download export_fig from the 
% MathWorks file exchange.

%% Setup
user= 'USER';
project_path = ['/home/' user '/chaste-source/StochasticPhenomenologicalModels'];
output_path = [ '/tmp/' user '/testoutput' ];
matlab_path = [ project_path '/Matlab' ];
% export_fig?
exportFig = 0;
% beats to investigate
beats = 1:49;
% conductivities to simulate
conductivities=[1 0.5 0.1 0.05];
% number of cell models to investigate
cell_models = 1:4;
num_cell_models = length(cell_models);
% initialise matrices to hold data
deterministic_mean = zeros(length(beats),length(conductivities), num_cell_models);
deterministic_variance = zeros(length(beats),length(conductivities), num_cell_models);
stochastic_mean = zeros(length(beats),length(conductivities), num_cell_models);
stochastic_variance = zeros(length(beats),length(conductivities), num_cell_models);
difference_variance = zeros(length(beats),length(conductivities), num_cell_models);
diff_max = zeros(length(beats),length(conductivities), num_cell_models);
diff_min = zeros(length(beats),length(conductivities), num_cell_models);
abs_diff = zeros(length(beats),length(conductivities), num_cell_models);

% open results text file
resultsFile = fopen('Fig2_Summary.dat', 'w');
% print header.
fprintf(resultsFile, 'Conductivities\n');
fprintf(resultsFile, '%1.2f ', conductivities);
fprintf('\n');

%% Load Data

% change into directory containing results.
cd( [ output_path '/StochasticsRegularPacing_archives' ] );

% loop over cell models
for cell_model = cell_models
    % loop over conductivities    
    for cond =1:length(conductivities)
        % loop over beats under investigation      	   
        for i=1:length(beats)
            % change into directory of current deterministic beat
            cd(['archive_CS_' num2str(conductivities(cond)) '_cell_model_' num2str(cell_model) '_stim_' num2str(beats(i)) '/']);
            
            % Get the data out of the relevant deterministic archive folder and reshape to square
            data = importdata('Apd_90_0pt40_Map.dat');
            deterministicdata = reshape(data,100,100);
            deterministicdata = deterministicdata';
            % remove boundaries to prevent boundary effects (see Cherry and Fenton 2011).
            deterministicdata = deterministicdata(:,25:75);
			cd ../

			% Repeat for the current stochastic beat
            cd(['archive_CS_' num2str(conductivities(cond)) '_cell_model_' num2str(cell_model) '_stim_' num2str(beats(i)) '_stochastic/']);
            
            % Get the data out of the relevant archive folder and reshape to square
            data = importdata('Apd_90_0pt40_Map.dat');
            stochasticdata = reshape(data,100,100);
            stochasticdata = stochasticdata';
            stochasticdata = stochasticdata(:,25:75);
            
			% calculate the difference between the deterministic and stochastic data
            difference = deterministicdata-stochasticdata;
            
            % now calculate stats

            % mean APDs in the tissue for this beat       		     
            stochastic_mean(i,cond, cell_model) = mean(mean(stochasticdata(2:end-1,2:end-1)));
            deterministic_mean(i,cond, cell_model) = mean(mean(deterministicdata(2:end-1,2:end-1)));
            % variance in APD in the tissue for this beat
            stochastic_variance(i,cond, cell_model) = var(reshape(stochasticdata(2:end-1,2:end-1), 98*49,1));
            deterministic_variance(i,cond, cell_model) = var(reshape(deterministicdata(2:end-1,2:end-1), 98*49,1));
            % variance in the difference in APD, by node, for this beat.		
            difference_vector = reshape(difference, 100*51,1);
            difference_variance(i,cond,cell_model) = var(difference_vector);
            % max and min differences
            diff_min(i,cond,cell_model) = min(min(difference));
            diff_max(i,cond,cell_model) = max(max(difference));
			% absolute differences
            abs_diff(i,cond, cell_model) = max(max(abs(difference)));
            
            cd ../
            
        end % beats      
    end % conductivities
end % cell models
    
%% Generate plots

% change into the Matlab directory
cd( matlab_path )

% Plot the mean APD for the stochastic and deterministic cases
% set colours and line markings 
det_colours = {'r--', 'g--', 'b--', 'k--'};  
stoch_colours = {'r', 'g', 'b', 'k'};
% loop over cell models
for cell_model = cell_models   
	% get the minimum and maximum of the data for this cell model over all conductivities and beats
    min_data = min([min(min(squeeze(deterministic_mean(:,:,cell_model)))), min(min(squeeze(stochastic_mean(:,:,cell_model))))]);
    max_data = max([max(squeeze(max(deterministic_mean(:,:,cell_model)))), max(max(squeeze(stochastic_mean(:,:,cell_model))))]);
    
	% plot the mean data per beat.
	% open figure
    figureHandle = figure('units','normalized','outerposition',[0.1 0.1 0.9 0.9]);
    hold on
	% loop over conductivities and plot deterministic and stochastic means
    for i=1:length(conductivities)
        det_colour = det_colours{i};
        plot(squeeze(deterministic_mean(:,i,cell_model)), det_colour, 'LineWidth', 3.0);
        stoch_colour = stoch_colours{i};
        plot(squeeze(stochastic_mean(:,i,cell_model)), stoch_colour, 'LineWidth', 3.0);
    end
	% increasing y axis slightly gives tidier plot.
    ylim([0.9975*min_data 1.0025*max_data]);
    xlim([0 50])
    % only label edge axes
    if (cell_model ==1 || cell_model == 3)
        ylabel('Mean APD (ms)')
    end
    if (cell_model == 3 || cell_model == 4)
        xlabel('Beat Number')
    end
    % get the axes in the figure   
    allAxesInFigure = findall(figureHandle,'type','axes');
    % cosmetics (font size, line width, turn off box)
    for ax=1:length(allAxesInFigure)
        set(allAxesInFigure(ax), 'FontSize', 30)
        set(findall(allAxesInFigure(ax), 'Type','text'), 'FontSize', 30)
        set(allAxesInFigure(ax), 'LineWidth', 3)
        set(allAxesInFigure(ax), 'Box', 'Off')
    end
    % if export_fig, export figure as eps.
    figName = (['Fig2_APDMean_cellmodel_' num2str(cell_model) '.eps']);
    if exportFig
    	export_fig(figName, '-eps', '-transparent');
	end
    
    % Plot the variance for the stochastic and determinisitic cases   
    colours = {'r', 'g', 'b', 'k'};
    % get the max of the data
    max_data = max([max(max(max(deterministic_variance(2:end,:,:)))), max(max(max(stochastic_variance(2:end,:,:))))]);
    % open figure
    figureHandle = figure('units','normalized','outerposition',[0.1 0.1 0.9 0.9]);
    subplot(1,2,1);
    hold on
	% loop over conductivities, plot the determininistic variance
    for i=1:length(conductivities)
        colour = colours{i};
        plot(squeeze(deterministic_variance(2:end,i,cell_model)), colour, 'LineWidth', 2.0);
    end
    ylim([0 1.05*max_data]);
    xlim([0 50])
    % label figure
    xlabel('Beat Number')
    ylabel('Variance in APD{_{90}} (ms{^2})')
    % loop over conductivities, plot the stochastic variance in new subplot
    subplot(1,2,2);
    hold on
    for i=1:length(conductivities)
        colour = colours{i};
        plot(squeeze(stochastic_variance(2:end,i,cell_model)), colour, 'LineWidth', 2.0);
    end
	% set axes
    ylim([0 1.05*max_data]);
    xlim([0 50])
    % label axes
    xlabel('Beat Number')
    ylabel('Variance in APD{_{90}} (ms{^2})')
    % do cosmetics
    allAxesInFigure = findall(figureHandle,'type','axes');
    
    for ax=1:length(allAxesInFigure)
        set(allAxesInFigure(ax), 'FontSize', 20)
        set(findall(allAxesInFigure(ax), 'Type','text'), 'FontSize', 26)
        set(allAxesInFigure(ax), 'LineWidth', 2)
        set(allAxesInFigure(ax), 'Box', 'Off')
    end
    
    figName = (['Fig2_APDVariance_cellmodel_' num2str(cell_model) '.eps']);
    if exportFig
    	export_fig(figName, '-eps', '-transparent');
	end   
       
    % Plot the maximum absolute difference for the stochastic and determinisitic cases
    
    colours = {'r', 'g', 'b', 'k'};
    % get the maximum of the data
    max_data = max(max(max(abs_diff)));
    % Open figure	
    figureHandle = figure('units','normalized','outerposition',[0.1 0.1 0.9 0.9]);
    % loop over conductivities and and plot the absolute difference.
	hold on
    for i=1:length(conductivities)
        colour = colours{i};
        plot(squeeze(abs_diff(:,i,cell_model)), colour, 'LineWidth', 3.0);
    end
    hold off
	% set axis limits
    max_data
    ylim([0 1.05*max_data]); %because min is negative
    xlim([0 50])
    % label edge axes only
    if (cell_model == 3 || cell_model == 4)
        xlabel('Beat Number')
    end
    
    if (cell_model == 1 || cell_model == 3)
        ylabel('Absolute Difference in APD{_{90}} (ms)')
    end
    % do cosmetics
    allAxesInFigure = findall(figureHandle,'type','axes');
    
    for ax=1:length(allAxesInFigure)
        set(allAxesInFigure(ax), 'FontSize', 30)
        set(findall(allAxesInFigure(ax), 'Type','text'), 'FontSize', 30)
        set(allAxesInFigure(ax), 'LineWidth', 3)
        set(allAxesInFigure(ax), 'Box', 'Off')
    end
    
    figName = (['Fig2_APDAbsoluteDifference_cellmodel_' num2str(cell_model) '.eps']);
    if exportFig
    	export_fig(figName, '-eps', '-transparent');
	end
    
    % write statistics to file

    % loop over conductivities
    for cond = 1:length(conductivities)   
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' deterministic mean (mean)\n']);
		fprintf(resultsFile, '%1.1f ', mean(squeeze(deterministic_mean(:,:,cell_model))));
		fprintf(resultsFile, '\n');
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' deterministic var (mean)\n']);
		fprintf(resultsFile, '%1.4f ', var(squeeze(deterministic_mean(:,:,cell_model))));
		fprintf(resultsFile, '\n');
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' stochastic mean (mean)\n']);
		fprintf(resultsFile, '%1.1f ', mean(squeeze(stochastic_mean(:,:,cell_model))));
		fprintf(resultsFile, '\n');
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' stochastic var (mean)\n']);
		fprintf(resultsFile, '%1.4f ', var(squeeze(stochastic_mean(:,:,cell_model))));
		fprintf(resultsFile, '\n');
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' minimum difference\n']);
		fprintf(resultsFile, '%1.3f ', min(squeeze(diff_min(:,:,cell_model))));
		fprintf(resultsFile, '\n');
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' maximum difference\n']);
		fprintf(resultsFile, '%1.3f ', max(squeeze(diff_min(:,:,cell_model))));
		fprintf(resultsFile, '\n');
		fprintf(resultsFile, ['Cell ' num2str(cell_model) ' maximum absolute difference\n']);
		fprintf(resultsFile, '%1.3f ', max(squeeze(abs_diff(:,:,cell_model))));
		fprintf(resultsFile, '\n');
    end
    
end % loop over cell models

% close text file.
fclose(resultsFile);

end
