/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "Walmsley2015.hpp"
#include <cmath>
#include <cassert>
#include <memory>
#include "Exception.hpp"
#include "OdeSystemInformation.hpp"
#include "RegularStimulus.hpp"
#include "HeartConfig.hpp"
#include "IsNan.hpp"
#include "RandomNumberGenerator.hpp"

double CellWalmsley2015FromCellML::Get_fast_inward_current__i_fi()
{
    return var_fast_inward_current__i_fi;
}

double CellWalmsley2015FromCellML::Get_membrane__C()
{
    return var_membrane__C;
}

double CellWalmsley2015FromCellML::Get_membrane__i_stim()
{
    return var_membrane__i_stim;
}

double CellWalmsley2015FromCellML::Get_slow_inward_current__i_si()
{
    return var_slow_inward_current__i_si;
}

double CellWalmsley2015FromCellML::Get_slow_outward_current__i_so()
{
    return var_slow_outward_current__i_so;
}

CellWalmsley2015FromCellML::CellWalmsley2015FromCellML(boost::shared_ptr<AbstractIvpOdeSolver> pSolver,
                                                       boost::shared_ptr<AbstractStimulusFunction> pIntracellularStimulus,
                                                       const double xLocation)
    : AbstractCardiacCell(
            pSolver,
            4,
            0,
            pIntracellularStimulus)
{
    //std::cout << "X loc = " << xLocation << "\n";
    // Time units: ms
    //
    this->mpSystemInfo = OdeSystemInformation<CellWalmsley2015FromCellML>::Instance();
    Init();
    // Standard parameters are for a Guinea Pig cell without stochasticity
    this->mParameters[0] = 0.2985;
    this->mParameters[1] = 0.1299;
    this->mParameters[2] = 0.2016;
    this->mParameters[3] =  0.006;
    this->mParameters[4] = 0.9502;
    this->mParameters[5] =   0.77;
    this->mParameters[6] = 0.6456;
    this->mParameters[7] = 1.5463;
    this->mParameters[8] = 0.0001;
    this->mParameters[9] = 0.0273;
    this->mParameters[10] = 462.44;
    this->mParameters[11] =  5.951;
    this->mParameters[12] = 2.6892;
    this->mParameters[13] = 1.9773;
    this->mParameters[14] = 39.317;
    this->mParameters[15] = 1.1926;
    this->mParameters[16] = 6;     // tau_w1minus
    this->mParameters[17] = 140;   // tau_w2minus
    this->mParameters[18] = 75.393;
    this->mParameters[19] = 10.026;
    this->mParameters[20] = 2.9338;
    this->mParameters[21] = 1.4622;
    this->mParameters[22] = 275.91;
    this->mParameters[23] = 0.0957;
    this->mParameters[24] =   0.95;
    this->mParameters[25] = 1.9465;
    this->mParameters[26] =  198.2;
    this->mParameters[27] = 2.2028;
    this->mParameters[28] = 94.055;
    this->mParameters[29] =    1.1;
    // Modify to render stochastic terms sigma_v and sigma_w settable.
    // this->mParameters[30] = 0.0012;//sigma_w
    // this->mParameters[31] = 0.0012;//sigma_s
}

CellWalmsley2015FromCellML::~CellWalmsley2015FromCellML()
{
}

void CellWalmsley2015FromCellML::VerifyStateVariables()
{

}

double CellWalmsley2015FromCellML::GetIIonic(const std::vector<double>* pStateVariables)
{
    if (!pStateVariables) pStateVariables = &rGetStateVariables();
    const std::vector<double>& rY = *pStateVariables;
    double var_membrane__V = rY[0];
    // Units: mV; Initial value: 0
    double var_fast_inward_current_v_gate__v = rY[1];
    // Units: dimensionless; Initial value: 1
    double var_slow_inward_current_w_gate__w = rY[2];
    // Units: dimensionless; Initial value: 1
    double var_slow_inward_current_s_gate__s = rY[3];
    // Units: dimensionless; Initial value: 0

    const double K = 1000;

    // Time constants
    const double tau_so = mParameters[14] +(mParameters[15] - mParameters[14])*(1+tanh(mParameters[25]*(var_membrane__V-mParameters[6])))/2;
    const double tau_o = (1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[3]))))*mParameters[10] + (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[3])))*mParameters[11];

    // Determine 'currents'
    const double J_fi = -var_fast_inward_current_v_gate__v*(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[0])))*(var_membrane__V-mParameters[0])*(mParameters[7]-var_membrane__V)/mParameters[23]; // 'fast inward' current
    const double J_so = (var_membrane__V-mParameters[8])*(1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1]))))/tau_o + (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1])))/tau_so; //'slow outward' current
    const double J_si = -(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1])))*var_slow_inward_current_w_gate__w*var_slow_inward_current_s_gate__s/mParameters[20]; // 'slow inward' current

    double i_ionic = J_fi + J_so + J_si;
    EXCEPT_IF_NOT(!std::isnan(i_ionic));
    return i_ionic;
}

void CellWalmsley2015FromCellML::EvaluateYDerivatives(double var_environment__time, const std::vector<double>& rY, std::vector<double>& rDY)
{
    // Inputs:
    // Time units: ms
    double var_membrane__V = rY[0];
    // Units: mV; Initial value: 0
    double var_fast_inward_current_v_gate__v = rY[1];
    // Units: dimensionless; Initial value: 1
    double var_slow_inward_current_w_gate__w = rY[2];
    // Units: dimensionless; Initial value: 1
    double var_slow_inward_current_s_gate__s = rY[3];
    // Units: dimensionless; Initial value: 0

    const double K = 1000;

    // Time constants
    const double tau_vminus = (1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[2]))))*mParameters[18] + (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[2])))*mParameters[19];
    const double tau_wminus = mParameters[16] + (mParameters[17] -mParameters[16])*(1+tanh(mParameters[26]*(var_slow_inward_current_w_gate__w-mParameters[5])))/2;
    const double tau_so = mParameters[14] +(mParameters[15] - mParameters[14])*(1+tanh(mParameters[25]*(var_membrane__V-mParameters[6])))/2;
    const double tau_s = (1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1]))))*mParameters[12] + (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1])))*mParameters[13];
    const double tau_o = (1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[3]))))*mParameters[10] + (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[3])))*mParameters[11];

    // Determine 'currents'
    const double J_fi = -var_fast_inward_current_v_gate__v*(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[0])))*(var_membrane__V-mParameters[0])*(mParameters[7]-var_membrane__V)/mParameters[23]; // 'fast inward' current
    const double J_so = (var_membrane__V-mParameters[8])*(1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1]))))/tau_o + (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1])))/tau_so; //'slow outward' current


    const double J_si = -(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1])))*var_slow_inward_current_w_gate__w*var_slow_inward_current_s_gate__s/mParameters[20]; // 'slow inward' current

    // calculate steady state values
    double v_inf = 0;
    if(var_membrane__V < mParameters[2])
    {
        v_inf = 1;
    }

    const double w_inf = (0.5 + 0.5*tanh(K*(var_slow_inward_current_w_gate__w-mParameters[29])))*
                         (1-(var_slow_inward_current_w_gate__w/mParameters[9]))
                         + (1-(0.5 + 0.5*tanh(K*(var_slow_inward_current_w_gate__w-mParameters[29]))))*mParameters[24];

    // Right hand sides of gating variables.
     double d_dt_v = ((1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[0]))))*(v_inf-var_fast_inward_current_v_gate__v)/tau_vminus -
            (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[0])))*var_fast_inward_current_v_gate__v/mParameters[21]);// + timestep_scaling*sigma_v*dW_v;


    double d_dt_w = ((1-(0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1]))))*(w_inf-var_slow_inward_current_w_gate__w)/tau_wminus -
            (0.5 + 0.5*tanh(K*(var_membrane__V-mParameters[1])))*var_slow_inward_current_w_gate__w/mParameters[22]);// + timestep_scaling*mParameters[30]*dW_w;

    double d_dt_s = (((1 + tanh(mParameters[27]*(var_membrane__V-mParameters[4])))/2 -var_slow_inward_current_s_gate__s)/tau_s);// + timestep_scaling*mParameters[31]*dW_s;

    // Right hand side of voltage equation
    double d_dt_membrane__V;
    if (mSetVoltageDerivativeToZero)
    {
        d_dt_membrane__V = 0.0;
    }
    else
    {
        var_membrane__C = 1.0; // uF_per_cm2
        var_membrane__i_stim = GetIntracellularAreaStimulus(var_environment__time);
        d_dt_membrane__V = ( -1.0 / var_membrane__C) * (J_fi + J_so + J_si + var_membrane__i_stim); // 'millivolt per millisecond'
    }

    rDY[0] = d_dt_membrane__V;
    rDY[1] = d_dt_v;
    rDY[2] = d_dt_w;
    rDY[3] = d_dt_s;

    EXCEPT_IF_NOT(!std::isnan(d_dt_membrane__V));
    EXCEPT_IF_NOT(!std::isnan(d_dt_v));
    EXCEPT_IF_NOT(!std::isnan(d_dt_w));
    EXCEPT_IF_NOT(!std::isnan(d_dt_s));
}


template<>
void OdeSystemInformation<CellWalmsley2015FromCellML>::Initialise(void)
{
    // Time units: ms
    //
    this->mSystemName = "Walmsley2015";

    this->mVariableNames.push_back("membrane_voltage");
    this->mVariableUnits.push_back("mV");
    this->mInitialConditions.push_back(0);

    this->mVariableNames.push_back("fast_inward_current_v_gate__v");
    this->mVariableUnits.push_back("dimensionless");
    this->mInitialConditions.push_back(1);

    this->mVariableNames.push_back("slow_inward_current_w_gate__w");
    this->mVariableUnits.push_back("dimensionless");
    this->mInitialConditions.push_back(0.775);

    this->mVariableNames.push_back("slow_inward_current_s_gate__s");
    this->mVariableUnits.push_back("dimensionless");
    this->mInitialConditions.push_back(0);

    this->mParameterNames.push_back("theta_v");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("theta_w");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("theta_vminus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("theta_o");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("u_s");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("w_wminus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("u_so");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("u_u");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("u_o");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_winf");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_o1");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_o2");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_s1");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_s2");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_so1");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_so2");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_w1minus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_w2minus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_v1minus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_v2minus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_si");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_vplus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_wplus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("tau_fi");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("w_infstar");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("k_so");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("k_wminus");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("k_s");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("V");
    this->mParameterUnits.push_back("dimensionless");

    this->mParameterNames.push_back("theta_wo");
    this->mParameterUnits.push_back("dimensionless");

//    this->mParameterNames.push_back("sigma_w");
//    this->mParameterUnits.push_back("dimensionless");
//
//    this->mParameterNames.push_back("sigma_s");
//    this->mParameterUnits.push_back("dimensionless");

    this->mInitialised = true;
}


// Serialization for Boost >= 1.36
#include "SerializationExportWrapperForCpp.hpp"
CHASTE_CLASS_EXPORT(CellWalmsley2015FromCellML)


