/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef TESTSINGLECELLSIMULATIONS_HPP_
#define TESTSINGLECELLSIMULATIONS_HPP_


#include <cxxtest/TestSuite.h>
#include <ctime>

#include <boost/shared_ptr.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/lexical_cast.hpp>

#include "Exception.hpp"

#include "EulerIvpOdeSolver.hpp"
#include "bueno_2007_endo.hpp"
#include "Walmsley2015_ParameterVector.hpp"
#include "Walmsley2015_ParameterVector_Stochastic.hpp"

#include "SimpleStimulus.hpp"
#include "RegularStimulus.hpp"
#include "PetscSetupAndFinalize.hpp"
#include "PostProcessingWriter.hpp"

#include "CommandLineArguments.hpp"
#include "CellProperties.hpp"

class TestSingleCellSimulations : public CxxTest::TestSuite
{
private:
	std::vector<double> mParameterVector;

public:
	/**
	 *
	 * Parameters can be defined at the top of this Test
	 */

	void TestModelVariables(void) throw (Exception)
    {
		double s_magnitude = -0.5; // Can make these model specific below. = -25.5;     // mV
		double s_duration = 2.0; // Can make these model specific below.  = 2.0 ;      // ms

		//double s_start = 10.0;           // ms
		double printing_time_step = 0.1;  // ms
		//double time_step = 0.00001; // ms

		double cycle_length = 1000.0;

		boost::shared_ptr<EulerIvpOdeSolver> p_solver(new EulerIvpOdeSolver);

		unsigned num_stims = 10u;

		std::string foldername = "SingleCellSims/BuenoOrovio2008";
		// Make above directories. This will NOT wipe an existing folder!
		OutputFileHandler steady_handler(foldername, false);

		// Overwritten!
		boost::shared_ptr<RegularStimulus> p_regular_stimulus(new RegularStimulus(s_magnitude, s_duration, cycle_length, 10));
		Cellbueno_2007_endoFromCellML* p_bo08_model = new Cellbueno_2007_endoFromCellML(p_solver, p_regular_stimulus);

		boost::shared_ptr<const AbstractOdeSystemInformation> p_bo08_ode_info = p_bo08_model->GetSystemInformation();
		std::string bo08_model_name = p_bo08_model->GetSystemName();
		unsigned u_index_bo08 = p_bo08_ode_info->GetStateVariableIndex("membrane_voltage");
		unsigned v_index_bo08 = p_bo08_ode_info->GetStateVariableIndex("fast_inward_current_v_gate__v");
		unsigned w_index_bo08 = p_bo08_ode_info->GetStateVariableIndex("slow_inward_current_w_gate__w");
		unsigned s_index_bo08 = p_bo08_ode_info->GetStateVariableIndex("slow_inward_current_s_gate__s");

		p_bo08_model -> SetStimulusFunction(p_regular_stimulus);

		// Set the max time step to be the printing time step (1ms).
		p_bo08_model->Compute(0, cycle_length*(double)(num_stims), printing_time_step );

		OdeSolution bo08_solution = p_bo08_model->Compute(0, cycle_length, printing_time_step );

		// Get variable properties
		std::vector<double> u_bo08 = bo08_solution.GetVariableAtIndex(u_index_bo08); // Voltage should always be zero
		std::vector<double> v_bo08 = bo08_solution.GetVariableAtIndex(v_index_bo08); // Voltage should always be zero
		std::vector<double> w_bo08 = bo08_solution.GetVariableAtIndex(w_index_bo08); // Voltage should always be zero
		std::vector<double> s_bo08 = bo08_solution.GetVariableAtIndex(s_index_bo08); // Voltage should always be zero
		std::vector<double> time_bo08 = bo08_solution.rGetTimes();

		out_stream bo08_data_traces_file;
		std::stringstream bo08_file_name_stream;
		bo08_file_name_stream << "DataTracesbo08_"<< cycle_length << "ms.dat";
		std::string bo08_data_traces_file_name = bo08_file_name_stream.str();
		bo08_data_traces_file = steady_handler.OpenOutputFile(bo08_data_traces_file_name);
		// Write solutions to disc
		for (unsigned i=0; i<u_bo08.size(); i++)
		{
			*bo08_data_traces_file << std::setprecision(4) <<
					time_bo08[i] << "\t" <<
					u_bo08[i] << "\t" <<
					v_bo08[i] << "\t" <<
					w_bo08[i] << "\t" <<
					s_bo08[i] << "\n";
		}
		*bo08_data_traces_file << std::flush;
		bo08_data_traces_file->close();
    }

	void TestStochasticsModel(void) throw(Exception)
	{
		double s_magnitude = -0.5; // Can make these model specific below. = -25.5;     // mV
		double s_duration = 2.0; // Can make these model specific below.  = 2.0 ;      // ms

		double s_start = 10.0;           // ms
		double printing_time_step = 0.1;  // ms
		//double time_step = 0.001; // ms

		boost::shared_ptr<EulerIvpOdeSolver> p_solver(new EulerIvpOdeSolver);

		unsigned num_stims = 100u;
		double cycle_length = 1000;

		std::string foldername = "SingleCellSims/Stochastics";
		// Make above directories. This will NOT wipe an existing folder!
		OutputFileHandler steady_handler(foldername, false);

		// Guinea Pig apex 1
		std::vector<double> gp_apex_1_params;
		gp_apex_1_params = boost::assign::list_of<double>
		(0.2985)(0.1299)(0.2016)(0.0060)(0.9502)(0.0160)(0.6456)(1.5463)(0.0001)(0.0273)(462.4436)(5.9510)(2.6892)(1.9773)(39.3173)(1.1926)(5.9208)(139.5427)(75.3934)(10.0256)(2.9338)(1.4622)(275.9146)(0.0957)(0.7847)(1.9465)(198.1980)(2.2028)(94.0553);

		// Guinea Pig apex 2
		std::vector<double> gp_apex_2_params;
		gp_apex_2_params = boost::assign::list_of<double>
		(0.3209)(0.1302)(0.1943)(0.0056)(0.9802)(0.0166)(0.6475)(1.6175)(-0.0008)(0.0410)(416.8280)(6.1000)(3.0565)(2.1917)(30.9108)(0.2569)(6.3650)(150.3936)(76.2596)(10.2019)(2.7270)(1.3532)(276.2747)(0.1094)(0.8426)(2.0059)(203.9794)(2.2932)(103.7665);

		// Guinea Pig base 1
		std::vector<double> gp_base_1_params;
		gp_base_1_params = boost::assign::list_of<double>
		(0.2749)(0.1254)(0.1908)(0.0058)(0.9212)(0.0154)(0.6645)(1.5920)(0.0004)(0.0360)(476.7330)(5.5794)(2.4822)(1.7640)(39.1266)(1.1066)(5.6937)(133.9741)(72.9930)(9.6374)(2.9016)(1.5253)(352.7995)(0.0950)(0.7252)(2.0227)(187.9975)(2.6529)(94.6066);

		// Guinea Pig base 2
		std::vector<double> gp_base_2_params;
		gp_base_2_params = boost::assign::list_of<double>
		(0.3005)(0.1202)(0.1714)(0.0062)(0.9391)(0.0175)(0.7072)(1.5749)(0.0003)(0.0290)(456.3476)(6.0416)(2.8133)(1.9486)(28.8126)(1.2874)(6.1828)(137.4365)(66.5924)(10.0310)(2.7844)(1.3333)(273.7935)(0.0944)(0.7848)(1.9425)(206.9896)(2.8119)(90.4975);

		// Above are parameter sets for the original BO08 model. These must now be adapted to the new model formulation

		gp_apex_1_params.push_back(1.1);
		gp_apex_1_params[5] = 0.77;
		gp_apex_1_params[16] = 80;
		gp_apex_1_params[17] = 100;
		gp_apex_1_params[24] = 0.95;
		gp_apex_1_params[30] = 0.0012;
		gp_apex_1_params[31] = 0.0012;

		gp_apex_2_params.push_back(1.1);
		gp_apex_2_params[5] = 0.77;
		gp_apex_2_params[16] = 80;
		gp_apex_2_params[17] = 100;
		gp_apex_2_params[24] = 1.00;
		gp_apex_2_params[30] = 0.0015;
		gp_apex_2_params[31] = 0.0020;

		gp_base_1_params.push_back(1.1);
		gp_base_1_params[5] = 0.77;
		gp_base_1_params[16] = 80;
		gp_base_1_params[17] = 100;
		gp_base_1_params[24] = 0.95;
		gp_base_1_params[30] = 0.0011;
		gp_base_1_params[31] = 0.0013;

		gp_base_2_params.push_back(1.1);
		gp_base_2_params[5] = 0.77;
		gp_base_2_params[16] = 80;
		gp_base_2_params[17] = 100;
		gp_base_2_params[24] = 0.95;
		gp_base_2_params[30] = 0.0011;
		gp_base_2_params[31] = 0.0012;

		std::vector<std::vector<double> > parameter_vectors;
		parameter_vectors.push_back(gp_apex_1_params);
		parameter_vectors.push_back(gp_apex_2_params);
		parameter_vectors.push_back(gp_base_1_params);
		parameter_vectors.push_back(gp_base_2_params);

		std::vector<std::string> parameter_set_names;
		parameter_set_names.push_back("GP_Apex_1_stoch");
		parameter_set_names.push_back("GP_Apex_2_stoch");
		parameter_set_names.push_back("GP_Base_1_stoch");
		parameter_set_names.push_back("GP_Base_2_stoch");

		boost::shared_ptr<RegularStimulus> p_regular_stimulus(new RegularStimulus(s_magnitude, s_duration, cycle_length, s_start));

		for (unsigned param_set =0u; param_set < parameter_vectors.size(); param_set++ )
		{
			CellWalmsley2015_ParameterVector_Stochastic* p_model = new CellWalmsley2015_ParameterVector_Stochastic(p_solver, p_regular_stimulus, parameter_vectors[param_set]);
			boost::shared_ptr<const AbstractOdeSystemInformation> p_ode_info = p_model->GetSystemInformation();
			unsigned u_index = p_ode_info->GetStateVariableIndex("membrane_voltage");

			p_model -> SetStimulusFunction(p_regular_stimulus);
			// Set the max time step to be the printing time step (1ms).
			for (unsigned n = 0u; n < num_stims; n++)
			{
				OdeSolution solution = p_model->Compute(0, cycle_length, printing_time_step );

				std::vector<double> u = solution.GetVariableAtIndex(u_index); // Voltage should always be zero
				std::vector<double> time = solution.rGetTimes(); // Voltage should always be zero

				out_stream data_trace_file;
				std::stringstream file_name_stream;
				file_name_stream << parameter_set_names[param_set]<<"_trace_"<< n << ".dat";
				data_trace_file = steady_handler.OpenOutputFile(file_name_stream.str());

				for (unsigned i=0; i<u.size(); i++)
				{
					*data_trace_file << time[i] << "\t" << u[i] << "\n";
				}
				*data_trace_file << std::flush;
				data_trace_file->close();
			}
		}

	}

	void TestStochasticsModelsPoincare(void) throw(Exception)
	{
		double s_magnitude = -0.5; // Can make these model specific below. = -25.5;     // mV
		double s_duration = 2.0; // Can make these model specific below.  = 2.0 ;      // ms

		double s_start = 10.0;           // ms
		double printing_time_step = 0.1;  // ms
		//double time_step = 0.001; // ms

		boost::shared_ptr<EulerIvpOdeSolver> p_solver(new EulerIvpOdeSolver);

		unsigned num_stims = 100u;

		std::string foldername = "SingleCellSims/PoincarePlots";
		// Make above directories. This will NOT wipe an existing folder!
		OutputFileHandler steady_handler(foldername, false);

		// Guinea Pig apex 1
		std::vector<double> gp_apex_1_params;
		gp_apex_1_params = boost::assign::list_of<double>
		(0.2985)(0.1299)(0.2016)(0.0060)(0.9502)(0.0160)(0.6456)(1.5463)(0.0001)(0.0273)(462.4436)(5.9510)(2.6892)(1.9773)(39.3173)(1.1926)(5.9208)(139.5427)(75.3934)(10.0256)(2.9338)(1.4622)(275.9146)(0.0957)(0.7847)(1.9465)(198.1980)(2.2028)(94.0553);

		// Guinea Pig apex 2
		std::vector<double> gp_apex_2_params;
		gp_apex_2_params = boost::assign::list_of<double>
		(0.3209)(0.1302)(0.1943)(0.0056)(0.9802)(0.0166)(0.6475)(1.6175)(-0.0008)(0.0410)(416.8280)(6.1000)(3.0565)(2.1917)(30.9108)(0.2569)(6.3650)(150.3936)(76.2596)(10.2019)(2.7270)(1.3532)(276.2747)(0.1094)(0.8426)(2.0059)(203.9794)(2.2932)(103.7665);

		// Guinea Pig base 1
		std::vector<double> gp_base_1_params;
		gp_base_1_params = boost::assign::list_of<double>
		(0.2749)(0.1254)(0.1908)(0.0058)(0.9212)(0.0154)(0.6645)(1.5920)(0.0004)(0.0360)(476.7330)(5.5794)(2.4822)(1.7640)(39.1266)(1.1066)(5.6937)(133.9741)(72.9930)(9.6374)(2.9016)(1.5253)(352.7995)(0.0950)(0.7252)(2.0227)(187.9975)(2.6529)(94.6066);

		// Guinea Pig base 2
		std::vector<double> gp_base_2_params;
		gp_base_2_params = boost::assign::list_of<double>
		(0.3005)(0.1202)(0.1714)(0.0062)(0.9391)(0.0175)(0.7072)(1.5749)(0.0003)(0.0290)(456.3476)(6.0416)(2.8133)(1.9486)(28.8126)(1.2874)(6.1828)(137.4365)(66.5924)(10.0310)(2.7844)(1.3333)(273.7935)(0.0944)(0.7848)(1.9425)(206.9896)(2.8119)(90.4975);

		// Above are parameter sets for the original BO08 model. These must now be adapted to the new model formulation

		gp_apex_1_params.push_back(1.1);
		gp_apex_1_params[5] = 0.77;
		gp_apex_1_params[16] = 80;
		gp_apex_1_params[17] = 100;
		gp_apex_1_params[24] = 0.95;
		gp_apex_1_params[30] = 0.0012;
		gp_apex_1_params[31] = 0.0012;

		gp_apex_2_params.push_back(1.1);
		gp_apex_2_params[5] = 0.77;
		gp_apex_2_params[16] = 80;
		gp_apex_2_params[17] = 100;
		gp_apex_2_params[24] = 1.00;
		gp_apex_2_params[30] = 0.0015;
		gp_apex_2_params[31] = 0.0020;

		gp_base_1_params.push_back(1.1);
		gp_base_1_params[5] = 0.77;
		gp_base_1_params[16] = 80;
		gp_base_1_params[17] = 100;
		gp_base_1_params[24] = 0.95;
		gp_base_1_params[30] = 0.0011;
		gp_base_1_params[31] = 0.0013;

		gp_base_2_params.push_back(1.1);
		gp_base_2_params[5] = 0.77;
		gp_base_2_params[16] = 80;
		gp_base_2_params[17] = 100;
		gp_base_2_params[24] = 0.95;
		gp_base_2_params[30] = 0.0011;
		gp_base_2_params[31] = 0.0012;

		std::vector<std::vector<double> > parameter_vectors;
		parameter_vectors.push_back(gp_apex_1_params);
		parameter_vectors.push_back(gp_apex_2_params);
		parameter_vectors.push_back(gp_base_1_params);
		parameter_vectors.push_back(gp_base_2_params);

		std::vector<std::string> parameter_set_names;
		parameter_set_names.push_back("GP_Apex_1");
		parameter_set_names.push_back("GP_Apex_2");
		parameter_set_names.push_back("GP_Base_1");
		parameter_set_names.push_back("GP_Base_2");

		std::vector<double> cycle_lengths = boost::assign::list_of(1000)(400)(300)(260)(220);

		unsigned param_set = 0u;
		//for (unsigned param_set =0u; param_set < parameter_vectors.size(); param_set++ )
		//{
			for (unsigned cl = 0u; cl < cycle_lengths.size(); cl++)
			{
				boost::shared_ptr<RegularStimulus> p_regular_stimulus(new RegularStimulus(s_magnitude, s_duration, cycle_lengths[cl], s_start));

				CellWalmsley2015_ParameterVector* p_model = new CellWalmsley2015_ParameterVector(p_solver, p_regular_stimulus, parameter_vectors[param_set]);
				boost::shared_ptr<const AbstractOdeSystemInformation> p_ode_info = p_model->GetSystemInformation();
				unsigned u_index = p_ode_info->GetStateVariableIndex("membrane_voltage");

				p_model -> SetStimulusFunction(p_regular_stimulus);
				// Set the max time step to be the printing time step (1ms).

				out_stream APD90_file;
				std::stringstream file_name_stream;
				file_name_stream << parameter_set_names[param_set]<<"_APD90_CL_"<< cycle_lengths[cl] << ".dat";
				APD90_file = steady_handler.OpenOutputFile(file_name_stream.str());
				std::cout << cycle_lengths[cl] << "\n";
				for (unsigned n = 0u; n < num_stims; n++)
				{
					OdeSolution solution = p_model->Compute(0, cycle_lengths[cl], printing_time_step );

					std::vector<double> u = solution.GetVariableAtIndex(u_index); // Voltage should always be zero
					CellProperties voltage_properties(u, solution.rGetTimes(),0.4);
					double apd90 = 0;

				    try
				    {
				    	apd90 = voltage_properties.GetLastActionPotentialDuration(90);
				    }
				    catch (const Exception &e)
				    {
				    	apd90 = -1;
				    }

					*APD90_file << apd90 << "\t" << std::flush;
				}
				APD90_file->close();
			}
		//}

		// Now do stochastics:
		gp_apex_1_params[30] = 0.0012;
		gp_apex_1_params[31] = 0.0012;

		gp_apex_2_params[30] = 0.0015;
		gp_apex_2_params[31] = 0.0020;

		gp_base_1_params[30] = 0.0011;
		gp_base_1_params[31] = 0.0013;

		gp_base_2_params[30] = 0.0011;
		gp_base_2_params[31] = 0.0012;

		std::vector<std::vector<double> > stochastic_parameter_vectors;
		stochastic_parameter_vectors.push_back(gp_apex_1_params);
		stochastic_parameter_vectors.push_back(gp_apex_2_params);
		stochastic_parameter_vectors.push_back(gp_base_1_params);
		stochastic_parameter_vectors.push_back(gp_base_2_params);

		//for (unsigned param_set =0u; param_set < parameter_vectors.size(); param_set++ )
		//{
			for (unsigned cl = 0u; cl < cycle_lengths.size(); cl++)
			{
				boost::shared_ptr<RegularStimulus> p_regular_stimulus(new RegularStimulus(s_magnitude, s_duration, cycle_lengths[cl], s_start));

				CellWalmsley2015_ParameterVector_Stochastic* p_model = new CellWalmsley2015_ParameterVector_Stochastic(p_solver, p_regular_stimulus, parameter_vectors[param_set]);
				boost::shared_ptr<const AbstractOdeSystemInformation> p_ode_info = p_model->GetSystemInformation();
				unsigned u_index = p_ode_info->GetStateVariableIndex("membrane_voltage");

				p_model -> SetStimulusFunction(p_regular_stimulus);
				// Set the max time step to be the printing time step (1ms).

				out_stream APD90_file;
				std::stringstream file_name_stream;
				file_name_stream << parameter_set_names[param_set]<<"_stoch_APD90_CL_"<< cycle_lengths[cl] << ".dat";
				APD90_file = steady_handler.OpenOutputFile(file_name_stream.str());
				std::cout << cycle_lengths[cl] << "\n";
				for (unsigned n = 0u; n < num_stims; n++)
				{
					OdeSolution solution = p_model->Compute(0, cycle_lengths[cl], printing_time_step );

					std::vector<double> u = solution.GetVariableAtIndex(u_index); // Voltage should always be zero
					CellProperties voltage_properties(u, solution.rGetTimes(),0.4);
					double apd90 = voltage_properties.GetLastActionPotentialDuration(90);
					*APD90_file << apd90 << "\t" << std::flush;
				}
				APD90_file->close();
			}
		//}

	}
};

#endif /* TESTSINGLECELLSIMULATIONS_HPP_ */

