/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef TESTSTOCHASTICCONVERGENCE_HPP_
#define TESTSTOCHASTICCONVERGENCE_HPP_


#include <cxxtest/TestSuite.h>
#include <ctime>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>

#include "AbstractCardiacCellWithModifiers.hpp"
#include "AbstractCardiacCellFactory.hpp"
#include "Exception.hpp"

#include "CellProperties.hpp"

#include "RegularStimulus.hpp"
#include "EulerIvpOdeSolver.hpp"
#include "Walmsley2015StochasticConvergenceTest.hpp"

#include "MonodomainProblem.hpp"
#include "PetscSetupAndFinalize.hpp"
#include "DistributedTetrahedralMesh.hpp"
#include "PostProcessingWriter.hpp"

#include "CommandLineArguments.hpp"

#include "Debug.hpp"

#include <sstream>

template <unsigned DIM>
class PointStimulusCellFactory : public AbstractCardiacCellFactory<DIM>
{
private:
    boost::shared_ptr<RegularStimulus> mpStimulus;
    double mAreaOrVolume;
    int mRefinementLevel;
    bool mFullTest;

public:
    PointStimulusCellFactory(const double& rStimMag, const double& rStimDuration, const double& rAreaOrVolume, const int& rRefinementLevel, const bool& rFullTest)
        : AbstractCardiacCellFactory<DIM>(),
          mpStimulus(new RegularStimulus(rStimMag, rStimDuration, 300, 1)),// Default stimulus introduced at t=0.
          mAreaOrVolume(rAreaOrVolume),
          mRefinementLevel(rRefinementLevel),
          mFullTest(rFullTest)

    {
        assert(DIM==1 || DIM==2);
    }

    AbstractCardiacCell* CreateCardiacCellForTissueNode(Node<DIM>* pNode)
    {
        double x = pNode->GetPoint()[0];
		double y = pNode->GetPoint()[1];

        AbstractCardiacCell* p_cell;
        if (x < 0.03)
        {
        	p_cell =  new CellWalmsley2015StochasticConvergenceTest(this->mpSolver, this->mpStimulus, mRefinementLevel, mFullTest, x, y);
        	// std::cout << "Cell with stimulus created\n" << std::flush;
        }
        else
        {
            p_cell =  new CellWalmsley2015StochasticConvergenceTest(this->mpSolver, this->mpZeroStimulus, mRefinementLevel, mFullTest, x, y);
        }
        return p_cell;
    }
};

class TestStochasticConvergence : public CxxTest::TestSuite
{

public:
    /*
     * This test will wipe $CHASTE_TEST_OUTPUT/StochasticConvergenceTest
     *
     * Parameters can be defined at the top of this Test
     */

    void TestStochSims() throw(Exception)
    {
        // Get command line arguments
        CommandLineArguments* p_args = CommandLineArguments::Instance();
        unsigned argc = *(p_args->p_argc); // has the number of arguments.
        std::cout << "# " << argc-1 << " arguments supplied.\n" << std::flush;

        // Define level of refinement. 1 is coarsest. Default is 1.

        int refinement_level = 1;

        bool Refinement_Level_Option = CommandLineArguments::Instance()->OptionExists("--refinement_level");
        if (Refinement_Level_Option == true) {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--refinement_level");
            refinement_level = atof(val);
        }

		if (refinement_level !=1 && refinement_level != 2 && refinement_level != 3 && refinement_level != 4 && refinement_level != 5 )
		{
			std::cerr << "refinement_level must take an integer value between 1 and 5 inclusive. Re-setting to 1 (coarsest).\n";
			refinement_level = 1;
		}
        std::cout << "refinement_level is "<< refinement_level << "\n";

        double print_time_step = 1.0;
        bool Print_Step_Option = CommandLineArguments::Instance()->OptionExists("--print_time_step");
        if (Print_Step_Option == true)
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--print_time_step");
            print_time_step = atof(val);
        }

        // The refinement level determines our ODE timestep. Base timestep is 0.01, to match the PDE timestep.
        double base_sde_timestep = 0.01;
        double timestep_scaling_factor;
        timestep_scaling_factor = pow(2, 1-refinement_level); //ie 1 if rl = 1, 1/2 if rl = 2 etc.
        double sde_timestep;
        sde_timestep = timestep_scaling_factor*base_sde_timestep;
        std::cout << sde_timestep << "\n";

        // Define conductivity scale

        double conductivity_scale = 1;

        bool Conductivity_Scale_Option = CommandLineArguments::Instance()->OptionExists("--conductivity_scale");
        if (Conductivity_Scale_Option == true) {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--conductivity_scale");
            conductivity_scale = atof(val);
        }

        std::cout << "conductivity_scale is "<< conductivity_scale << "\n";

        double sheet_width = 0.2;

        bool Sheet_Width_Option = CommandLineArguments::Instance()->OptionExists("--width");
        if (Sheet_Width_Option == true) {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--width");
            sheet_width = atof(val);
        }

        std::cout << "sheet_width is "<< sheet_width << "\n";

        // Define whether we use the full set of paths or just use the test paths to ensure the file works. Default is test.

        // Define conductivity scale

        bool full_test = false;

        bool Full_Test_Option = CommandLineArguments::Instance()->OptionExists("--full_test");
        if (Full_Test_Option == true) {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--full_test");
            full_test = atof(val);
        }

        std::cout << "full_test is "<< full_test << "\n";

        // If we are doing the full test, we must first unzip the data. ~1.6gigs zipped, ~4 gigs unzipped.
        // Declare this in either case, only used if full_test is true
		if (full_test==true)
		{
			char BrownianPathsDatasetName[1024];
			sprintf(BrownianPathsDatasetName, "projects/StochasticPhenomenologicalModels/test/data/Local/full_brownian_path/");
			std::string dataset_name;
			dataset_name.assign(BrownianPathsDatasetName);
			FileFinder file(dataset_name, RelativeTo::ChasteSourceRoot);
			std::cout << dataset_name << "\n";
			if(!file.Exists())
			{
				EXCEPTION("No data folder at " << file.GetAbsolutePath());
			}
		}

        DistributedTetrahedralMesh<2,2> mesh;
        double h=0.01; //cm
        unsigned num_elem_x = (unsigned)(sheet_width/h);  // num elements to make 0.2 cm
        unsigned num_elem_y = (unsigned)(sheet_width/h);  // num elements to make 0.2 cm
        double stim_mag = -500;
        double stim_dur = 3;
        double area = 0.005;

        mesh.ConstructRectangularMesh(num_elem_x, num_elem_y);
        mesh.Scale(h,h); // Get mesh into units of cm.

        // if just testing code, duration is 1.0ms
        HeartConfig::Instance()->SetSimulationDuration(1.0); //ms
        // otherwise, duration is 250ms.
        if (full_test==true)
        {
        	HeartConfig::Instance()->SetSimulationDuration(250.0); //ms
        }
		
		// Simulation config
        std::stringstream out_dir;
        out_dir << std::setprecision(3) << "StochasticConvergenceTest/RefinementLevel_" << refinement_level<<
        		"_PrintTimeStep_" << print_time_step << "_CondScale_" << conductivity_scale;
        HeartConfig::Instance()->SetOutputDirectory(out_dir.str());
        HeartConfig::Instance()->SetOutputFilenamePrefix("results");
        HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(sde_timestep,0.01,print_time_step);
        HeartConfig::Instance()->SetIntracellularConductivities(Create_c_vector(1.4*conductivity_scale*1.171,1.4*conductivity_scale*1.171));
        HeartConfig::Instance()->SetSurfaceAreaToVolumeRatio(1400.0); // 1/cm
        HeartConfig::Instance()->SetCapacitance(1.0); // uF/cm^2

        // Postprocessing options
        std::vector<std::pair<double,double> > apds_requested;
        apds_requested.push_back(std::pair<double, double>(90,0.4)); //repolarisation percentage and threshold
        apds_requested.push_back(std::pair<double, double>(50,0.4)); //repolarisation percentage and threshold
        HeartConfig::Instance()->SetApdMaps(apds_requested);
		
		// Set and run simulations
        PointStimulusCellFactory<2> cell_factory(stim_mag, stim_dur, area, refinement_level, full_test);
        MonodomainProblem<2> monodomain_problem( &cell_factory );
        monodomain_problem.SetMesh(&mesh);
        monodomain_problem.Initialise();
        monodomain_problem.Solve();

    }

};

#endif /* TESTSTOCHASTICCONVERGENCE_HPP_ */

