/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef TESTSTOCHASTICRANDOMCELLREGULARPACING_HPP_
#define TESTSTOCHASTICRANDOMCELLREGULARPACING_HPP_


#include <cxxtest/TestSuite.h>
#include <ctime>
#include <boost/lexical_cast.hpp>
#include <boost/assign/list_of.hpp>

#include "CheckpointArchiveTypes.hpp" // Needs to be before other Chaste code
#include "CardiacSimulationArchiver.hpp"
#include "ArchiveOpener.hpp"
#include "Exception.hpp"
#include "DistributedVector.hpp"
#include "DistributedVectorFactory.hpp"

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/shared_ptr.hpp>

//#include "AbstractCardiacCell.hpp"
//#include "AbstractCvodeCell.hpp"
#include "AbstractCardiacCellWithModifiers.hpp"
#include "AbstractCardiacCellFactory.hpp"
//#include "Exception.hpp"

#include "CellProperties.hpp"

#include "SimpleStimulus.hpp"
#include "RegularStimulus.hpp"
#include "EulerIvpOdeSolver.hpp"
#include "Walmsley2015_ParameterVector.hpp"
#include "Walmsley2015_ParameterVector_Stochastic.hpp"
//#include "StimProtocol2D_Regular.hpp"

#include "MonodomainProblem.hpp"
#include "PetscSetupAndFinalize.hpp"
#include "DistributedTetrahedralMesh.hpp"
#include "TrianglesMeshReader.hpp"
#include "PostProcessingWriter.hpp"

#include "CommandLineArguments.hpp"

//#include "Debug.hpp"
#include "Exception.hpp"

#include <sstream>

std::vector<unsigned> ReadCellNumberFile(std::string fileName)
{
	std::vector<unsigned> cell_number_vector;
	std::ifstream indata; // indata is like cin
	indata.open(fileName.c_str()); // opens the file
	if(!indata.good())
	{ // file couldn't be opened
		EXCEPTION("Couldn't open data file: " + fileName);
	}

	while (indata.good())
	{
		std::string this_line;
		getline(indata, this_line);

		if (this_line=="" || this_line=="\r")
		{
			if (indata.eof())
			{    // If the blank line is the last line carry on OK.
				break;
			}
			else
			{
				EXCEPTION("No data found on this line");
			}
		}
		std::stringstream line(this_line);

		unsigned cell_number;
		while(line >> cell_number)
		{
			cell_number_vector.push_back(cell_number);
		}
	}
	return cell_number_vector;

}

template<class CELLMODEL, unsigned ELEMENT_DIM, unsigned SPACE_DIM = ELEMENT_DIM>
class RegularStimRandomCellFactory : public AbstractCardiacCellFactory<ELEMENT_DIM, SPACE_DIM>
{
private:
    boost::shared_ptr<RegularStimulus> mpStimulus;
    std::vector<unsigned> mCellModelNumbers;
    std::vector<std::vector<double> > mParameterVectors;
    double mAreaOrVolume;

public:
    RegularStimRandomCellFactory(const double& rStimMag, const double& rStimDuration, const double& rAreaOrVolume, const double& rCycleLength, const std::vector<std::vector<double> >& rParameterVectors, std::vector<unsigned> rCellModelNumbers)
        : AbstractCardiacCellFactory<ELEMENT_DIM, SPACE_DIM>(),
          mParameterVectors(rParameterVectors),
          mAreaOrVolume(rAreaOrVolume)
    {
	    assert(ELEMENT_DIM==1 || ELEMENT_DIM==2);
	    assert(rCycleLength > 0);

	    mCellModelNumbers = rCellModelNumbers;

	    boost::shared_ptr<RegularStimulus> p_regular_stim(new RegularStimulus(rStimMag,rStimDuration, rCycleLength, 0));

	    mpStimulus = p_regular_stim;

   }

    AbstractCardiacCell* CreateCardiacCellForTissueNode(Node<SPACE_DIM>* pNode)
    {
        double x = pNode->GetPoint()[0];
	unsigned nodeIndex = pNode->GetIndex();

        unsigned cell_model_number = mCellModelNumbers[nodeIndex];

        switch(cell_model_number)
        {
        case 0u:
        case 1u:
        case 2u:
        case 3u:
        if (x < 0.03)
        {
        	return new CELLMODEL(this->mpSolver, this->mpStimulus, this->mParameterVectors[cell_model_number]);
        }
        else
        {
            return new CELLMODEL(this->mpSolver, this->mpZeroStimulus, this->mParameterVectors[cell_model_number]);
        }
        break;

        default:
        	std::stringstream exception_stream;
        	exception_stream << "Incorrect cell number " << cell_model_number << " at nodeIndex " << nodeIndex << "\n";
        	EXCEPTION(exception_stream.str());

        }
    }
};

class TestStochasticRandomCellRegularPacing : public CxxTest::TestSuite
{

public:


    void Test2DSimulations() throw(Exception)
    {
        // Get command line arguments
        CommandLineArguments* p_args = CommandLineArguments::Instance();
        unsigned argc = *(p_args->p_argc); // has the number of arguments.
        if (PetscTools::AmMaster())
        {
        	std::cout << "# " << argc-1 << " arguments supplied.\n" << std::flush;
        }

        // Define conductivity scale
        double conductivity_scale = 1;
        if (CommandLineArguments::Instance()->OptionExists("--conductivity_scale"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--conductivity_scale");
            conductivity_scale = atof(val);
        }

        // Define spatial discretisation for finite element mesh
        double h = 0.01; // cm
        if (CommandLineArguments::Instance()->OptionExists("--spatial_discretization"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--spatial_discretization");
            h = atof(val);
        }

        // Define ode time step
        double ode_time_step = 0.005; //ms
        if (CommandLineArguments::Instance()->OptionExists("--ode_time_step"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--ode_time_step");
            ode_time_step = atof(val);
        }

        // Define pde time step
        double pde_time_step = 0.01; //ms
        if (CommandLineArguments::Instance()->OptionExists("--pde_time_step"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--pde_time_step");
            pde_time_step = atof(val);
        }

        // Set number of stimuli to apply
        unsigned num_stims = 5;
        if (CommandLineArguments::Instance()->OptionExists("--num_stims"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--num_stims");
            num_stims = atof(val);
        }
        // Check at least one stimulus is applied
        assert(num_stims >= 1);

        // Set cycle length in ms
        unsigned cycle_length = 250;
        if (CommandLineArguments::Instance()->OptionExists("--cycle_length"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--cycle_length");
            cycle_length = atof(val);
        }
        // Check that cycle length > 0
        assert(cycle_length > 0.0);

        std::string load_file_path = "";
        if (CommandLineArguments::Instance()->OptionExists("--load_file"))
        {
        	load_file_path = CommandLineArguments::Instance()->GetStringCorrespondingToOption("--load_file");
        }

        if (PetscTools::AmMaster())
        {
        	std::cout << "conductivity_scale is "<< conductivity_scale << "\n";
        	std::cout << "spatial_discretization is "<< h << "\n";
        	std::cout << "ode_time_step is "<< ode_time_step << "\n";
        	std::cout << "pde_time_step is "<< pde_time_step << "\n";
        	std::cout << "num_stims is "<< num_stims << "\n";
        	std::cout << "load_file_path is " << load_file_path << "\n";
        }

        // Create the mesh

        DistributedTetrahedralMesh<2,2> mesh;
        unsigned num_elem_x = (unsigned)(1.0/h);  // num elements to make 1cm
        unsigned num_elem_y = (unsigned)(1.0/h);  // num elements to make 1cm
        double stim_mag = -500;
        double stim_dur = 3;
        double area = 0.005;

        mesh.ConstructRectangularMesh(num_elem_x+1, num_elem_y+1);
        mesh.Scale(h,h); // Get mesh into units of cm.

        // Generate the cell model numbers
        unsigned num_nodes = mesh.GetNumNodes();
        std::vector<unsigned> cell_numbers_vector;

        // If option exists, load from file:
        if(CommandLineArguments::Instance()->OptionExists("--load_file"))
        {
        	FileFinder cell_number_file_finder(load_file_path, RelativeTo::ChasteSourceRoot);
        	if(!cell_number_file_finder.Exists())
        	{
        		EXCEPTION("Cannot find cell number file at "+load_file_path);
        	}

        	cell_numbers_vector = ReadCellNumberFile(cell_number_file_finder.GetAbsolutePath());
        }

        // Else, generate the cell numbers
        else
        {
        	if (PetscTools::AmMaster())
        	{
        		std::cout << "Generating a cell number distribution\n";
        	}

        	double uniform_random_number;
        	double random_number;
        	double max_random_number;
        	unsigned current_cell_number;
        	srand((unsigned)time(NULL));
        	for (unsigned i=0u; i<num_nodes ; i++)
        	{
        		random_number = (double) rand();
        		max_random_number = (double) (RAND_MAX) + 1;
        	    uniform_random_number= random_number / max_random_number;
        	    current_cell_number = boost::lexical_cast<unsigned> (floor(4.0*uniform_random_number));
        	    //std::cout << current_cell_number << "\n";
        	    cell_numbers_vector.push_back(current_cell_number);
        	}
        }

        if (cell_numbers_vector.size() != num_nodes)
        {
        	if (PetscTools::AmMaster())
        	{
        		std::cout << "Cell numbers vector length: " << cell_numbers_vector.size() << "\n";
        		std::cout << "Number of nodes in mesh: " << num_nodes << "\n";
        	}
        	EXCEPTION("Number of elements in mesh does not match number of cell model numbers provided");
        }

        // Cell model parameters
		// Guinea Pig apex 1
		std::vector<double> gp_apex_1_params;
		gp_apex_1_params = boost::assign::list_of<double>
		(0.2985)(0.1299)(0.2016)(0.0060)(0.9502)(0.0160)(0.6456)(1.5463)(0.0001)(0.0273)(462.4436)(5.9510)(2.6892)(1.9773)(39.3173)(1.1926)(5.9208)(139.5427)(75.3934)(10.0256)(2.9338)(1.4622)(275.9146)(0.0957)(0.7847)(1.9465)(198.1980)(2.2028)(94.0553);

    	// Guinea Pig apex 2
		std::vector<double> gp_apex_2_params;
		gp_apex_2_params = boost::assign::list_of<double>
		(0.3209)(0.1302)(0.1943)(0.0056)(0.9802)(0.0166)(0.6475)(1.6175)(-0.0008)(0.0410)(416.8280)(6.1000)(3.0565)(2.1917)(30.9108)(0.2569)(6.3650)(150.3936)(76.2596)(10.2019)(2.7270)(1.3532)(276.2747)(0.1094)(0.8426)(2.0059)(203.9794)(2.2932)(103.7665);

		// Guinea Pig base 1
		std::vector<double> gp_base_1_params;
		gp_base_1_params = boost::assign::list_of<double>
		(0.2749)(0.1254)(0.1908)(0.0058)(0.9212)(0.0154)(0.6645)(1.5920)(0.0004)(0.0360)(476.7330)(5.5794)(2.4822)(1.7640)(39.1266)(1.1066)(5.6937)(133.9741)(72.9930)(9.6374)(2.9016)(1.5253)(352.7995)(0.0950)(0.7252)(2.0227)(187.9975)(2.6529)(94.6066);

		// Guinea Pig base 2
		std::vector<double> gp_base_2_params;
		gp_base_2_params = boost::assign::list_of<double>
		(0.3005)(0.1202)(0.1714)(0.0062)(0.9391)(0.0175)(0.7072)(1.5749)(0.0003)(0.0290)(456.3476)(6.0416)(2.8133)(1.9486)(28.8126)(1.2874)(6.1828)(137.4365)(66.5924)(10.0310)(2.7844)(1.3333)(273.7935)(0.0944)(0.7848)(1.9425)(206.9896)(2.8119)(90.4975);

		// Make into 'new' model:
		gp_apex_1_params.push_back(1.1);
		gp_apex_1_params[5] = 0.77;
		gp_apex_1_params[16] = 80;
		gp_apex_1_params[17] = 100;
		gp_apex_1_params[24] = 0.95;

		gp_apex_2_params.push_back(1.1);
		gp_apex_2_params[5] = 0.77;
		gp_apex_2_params[16] = 80;
		gp_apex_2_params[17] = 100;
		gp_apex_2_params[24] = 1.00;

		gp_base_1_params.push_back(1.1);
		gp_base_1_params[5] = 0.77;
		gp_base_1_params[16] = 80;
		gp_base_1_params[17] = 100;
		gp_base_1_params[24] = 0.95;

		gp_base_2_params.push_back(1.1);
		gp_base_2_params[5] = 0.77;
		gp_base_2_params[16] = 80;
		gp_base_2_params[17] = 100;
		gp_base_2_params[24] = 0.95;

		std::vector<std::vector<double> > parameter_vectors;
		parameter_vectors.push_back(gp_apex_1_params);
		parameter_vectors.push_back(gp_apex_2_params);
		parameter_vectors.push_back(gp_base_1_params);
		parameter_vectors.push_back(gp_base_2_params);


		std::stringstream out_dir_stream;
		out_dir_stream << "RandomCellModel_CS_" << conductivity_scale;

        std::string archive_dir_base(out_dir_stream.str()+"_archives/");
        std::string archive_dir_current;

        // Setup
	HeartConfig::Instance()->SetVisualizeWithMeshalyzer();
        HeartConfig::Instance()->SetSimulationDuration(cycle_length); //ms
        HeartConfig::Instance()->SetOutputDirectory(out_dir_stream.str());
        HeartConfig::Instance()->SetOutputFilenamePrefix("results");
        HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(ode_time_step,pde_time_step,0.1);
        HeartConfig::Instance()->SetIntracellularConductivities(Create_c_vector(1.4*conductivity_scale*1.171, 1.4*conductivity_scale*1.171));
        HeartConfig::Instance()->SetSurfaceAreaToVolumeRatio(1400.0); // 1/cm
        HeartConfig::Instance()->SetCapacitance(1.0); // uF/cm^2

        std::vector<std::pair<double,double> > apds_requested;
        apds_requested.push_back(std::pair<double, double>(90,0.4)); //repolarisation percentage and threshold
        HeartConfig::Instance()->SetApdMaps(apds_requested);
        std::vector<double> excitation_threshold;
        excitation_threshold.push_back(double (0.4));
        HeartConfig::Instance()->SetUpstrokeTimeMaps(excitation_threshold);
        HeartConfig::Instance()->SetMaxUpstrokeVelocityMaps(excitation_threshold);

        for (unsigned stim_counter=0; stim_counter < num_stims; stim_counter++ )
        {
            // Load problem
            MonodomainProblem<2> *p_monodomain_problem;
            if (stim_counter==0)
            {
            	RegularStimRandomCellFactory<CellWalmsley2015_ParameterVector,2> cell_factory(stim_mag, stim_dur, area, cycle_length, parameter_vectors, cell_numbers_vector);
                p_monodomain_problem = new MonodomainProblem<2>( &cell_factory );
                p_monodomain_problem->SetMesh(&mesh);
                p_monodomain_problem->Initialise();

                OutputFileHandler steady_handler(archive_dir_base, false);
            	out_stream cell_number_save_file;
            	cell_number_save_file = steady_handler.OpenOutputFile("CellModelNumbers.dat");
            	for (unsigned cell_no = 0u; cell_no < cell_numbers_vector.size(); cell_no++)
            	{
            		*cell_number_save_file << cell_numbers_vector[cell_no] << "\n";
            	}
            	cell_number_save_file->close();
            }
            else
            {
                p_monodomain_problem = CardiacSimulationArchiver<MonodomainProblem<2> >::Load(archive_dir_current);
            }

            HeartConfig::Instance()->SetSimulationDuration((double) (stim_counter+1)*cycle_length); //ms

            // set new directories to work from
            std::stringstream stringoutput;
            stringoutput << "stim_" << stim_counter;
            std::string stim_counter_string = stringoutput.str();

            archive_dir_current = archive_dir_base + "archive_" + stim_counter_string;
            HeartConfig::Instance()->SetOutputFilenamePrefix("results_" + stim_counter_string);

            // Solve problem (this does the postprocessing too when HeartConfig options are set).
            p_monodomain_problem->Solve();

            // Save problem to archive
            CardiacSimulationArchiver<MonodomainProblem<2> >::Save(*p_monodomain_problem, archive_dir_current, false);
            std::cout << "Archived to " << archive_dir_current << "\n" << std::flush;

            if (PetscTools::AmMaster()) // Best to have only the master processor messing with files
            {
                // Copy the postprocessing results into the archive folders so they aren't wiped.
                std::vector<std::string> files;
                files.push_back("Apd_90_0pt40_Map");
                files.push_back("MaxUpstrokeVelocityMap_0pt40");
                files.push_back("UpstrokeTimeMap_0pt40");
                for (unsigned i=0; i<files.size(); i++)
                {
                    std::string command = "mv " +  OutputFileHandler::GetChasteTestOutputDirectory() + HeartConfig::Instance()->GetOutputDirectory() + "/output/" + files[i] + ".dat "
                                          + OutputFileHandler::GetChasteTestOutputDirectory() + archive_dir_current + "/" + files[i] + ".dat";

                    // If we are in an PetscTools::AmMaster() block we need to use this instead of EXPECT0.
                    ABORT_IF_NON0(system, command);
                }
            }
            PetscTools::Barrier();

        }// close for loop

        // Now do the stochastic version

		gp_apex_1_params.push_back(0.0012);
		gp_apex_1_params.push_back(0.0012);

		gp_apex_2_params.push_back(0.0015);
		gp_apex_2_params.push_back(0.0020);

		gp_base_1_params.push_back(0.0011);
		gp_base_1_params.push_back(0.0013);

		gp_base_2_params.push_back(0.0011);
		gp_base_2_params.push_back(0.0012);

		std::vector<std::vector<double> > stochastic_parameter_vectors;
		stochastic_parameter_vectors.push_back(gp_apex_1_params);
		stochastic_parameter_vectors.push_back(gp_apex_2_params);
		stochastic_parameter_vectors.push_back(gp_base_1_params);
		stochastic_parameter_vectors.push_back(gp_base_2_params);

		out_dir_stream <<"_stochastic";

        std::string stochastic_archive_dir_base(out_dir_stream.str()+"_archives/");
        std::string stochastic_archive_dir_current;

        HeartConfig::Instance()->SetOutputDirectory(out_dir_stream.str());

        for (unsigned stim_counter=0; stim_counter < num_stims; stim_counter++ )
        {
            // Load problem
            MonodomainProblem<2> *p_monodomain_problem;
            if (stim_counter==0)
            {
            	RegularStimRandomCellFactory<CellWalmsley2015_ParameterVector_Stochastic,2> cell_factory(stim_mag, stim_dur, area, cycle_length, stochastic_parameter_vectors, cell_numbers_vector);
                p_monodomain_problem = new MonodomainProblem<2>( &cell_factory );
                p_monodomain_problem->SetMesh(&mesh);
                p_monodomain_problem->Initialise();

                //Write the cell numbers used to a file
                OutputFileHandler steady_handler(stochastic_archive_dir_base, false);
            	out_stream cell_number_save_file;
            	cell_number_save_file = steady_handler.OpenOutputFile("CellModelNumbers.dat");
            	for (unsigned cell_no = 0u; cell_no < cell_numbers_vector.size(); cell_no++)
            	{
            		*cell_number_save_file << cell_numbers_vector[cell_no] << "\n";
            	}
            	cell_number_save_file->close();
            }
            else
            {
                p_monodomain_problem = CardiacSimulationArchiver<MonodomainProblem<2> >::Load(stochastic_archive_dir_current);
            }

            HeartConfig::Instance()->SetSimulationDuration((double) (stim_counter+1)*cycle_length); //ms

            // set new directories to work from
            std::stringstream stringoutput;
            stringoutput << "stim_" << stim_counter;
            std::string stim_counter_string = stringoutput.str();

            stochastic_archive_dir_current = stochastic_archive_dir_base + "archive_" + stim_counter_string;
            HeartConfig::Instance()->SetOutputFilenamePrefix("results_" + stim_counter_string);

            // Solve problem (this does the postprocessing too when HeartConfig options are set).
            p_monodomain_problem->Solve();

            // Save problem to archive
            CardiacSimulationArchiver<MonodomainProblem<2> >::Save(*p_monodomain_problem, stochastic_archive_dir_current, false);
            std::cout << "Archived to " << stochastic_archive_dir_current << "\n" << std::flush;

            if (PetscTools::AmMaster()) // Best to have only the master processor messing with files
            {
                // Copy the postprocessing results into the archive folders so they aren't wiped.
                std::vector<std::string> files;
                files.push_back("Apd_90_0pt40_Map");
                files.push_back("MaxUpstrokeVelocityMap_0pt40");
                files.push_back("UpstrokeTimeMap_0pt40");
                for (unsigned i=0; i<files.size(); i++)
                {
                    std::string command = "mv " +  OutputFileHandler::GetChasteTestOutputDirectory() + HeartConfig::Instance()->GetOutputDirectory() + "/output/" + files[i] + ".dat "
                                          + OutputFileHandler::GetChasteTestOutputDirectory() + stochastic_archive_dir_current + "/" + files[i] + ".dat";

                    // If we are in an PetscTools::AmMaster() block we need to use this instead of EXPECT0.
                    ABORT_IF_NON0(system, command);
                }
            }
            PetscTools::Barrier();

        }// close for loop

        // Report time spent on simulations
        HeartEventHandler::Headings();
        HeartEventHandler::Report();
    }//close void Test2dSimulations

};

#endif /* TESTSTOCHASTICRANDOMCELLREGULARPACING_HPP_ */

