/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef TESTSTOCHASTICSRAPIDPACING_HPP_
#define TESTSTOCHASTICSRAPIDPACING_HPP_


#include <cxxtest/TestSuite.h>
#include <ctime>
#include <boost/lexical_cast.hpp>
#include <boost/assign/list_of.hpp>

#include "CheckpointArchiveTypes.hpp" // Needs to be before other Chaste code
#include "CardiacSimulationArchiver.hpp"
#include "ArchiveOpener.hpp"
#include "Exception.hpp"
#include "DistributedVector.hpp"
#include "DistributedVectorFactory.hpp"

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/shared_ptr.hpp>

#include "AbstractCardiacCellWithModifiers.hpp"
#include "AbstractCardiacCellFactory.hpp"

#include "CellProperties.hpp"

#include "SimpleStimulus.hpp"
#include "MultiStimulus.hpp"
#include "EulerIvpOdeSolver.hpp"
#include "Walmsley2015_ParameterVector.hpp"
#include "Walmsley2015_ParameterVector_Stochastic.hpp"//

#include "MonodomainProblem.hpp"
#include "PetscSetupAndFinalize.hpp"
#include "DistributedTetrahedralMesh.hpp"
#include "PostProcessingWriter.hpp"

#include "CommandLineArguments.hpp"

template<class CELLMODEL, unsigned ELEMENT_DIM, unsigned SPACE_DIM = ELEMENT_DIM>
class RapidPacingCellFactory : public AbstractCardiacCellFactory<ELEMENT_DIM, SPACE_DIM>
{
private:
    boost::shared_ptr<MultiStimulus> mpStimulus;
    std::vector<double> mParameterVector;
    double mAreaOrVolume;

public:
    RapidPacingCellFactory(const double& rStimMag, const double& rStimDuration, const double& rAreaOrVolume, const double& rInitialCycleLength, const double& rCycleLengthReduction, const unsigned& rNumStims, const unsigned& rNumCycleLengths, const std::vector<double>& rParameterVector)
        : AbstractCardiacCellFactory<ELEMENT_DIM, SPACE_DIM>(),
          mParameterVector(rParameterVector),
          mAreaOrVolume(rAreaOrVolume)
    {
	    assert(ELEMENT_DIM==1 || ELEMENT_DIM==2);
	    assert(rInitialCycleLength > 0);

	    boost::shared_ptr<MultiStimulus> p_multi_stim(new MultiStimulus());

	    double start_time=0;
	    double current_cycle_length = rInitialCycleLength;

	    for (unsigned cl_counter=0; cl_counter < rNumCycleLengths; cl_counter++)
		{
			for(unsigned stim_counter = 0; stim_counter < rNumStims; stim_counter++ )
			{
				boost::shared_ptr<SimpleStimulus> p_init_stim(new SimpleStimulus(rStimMag,rStimDuration,(double) stim_counter* current_cycle_length + start_time));
				p_multi_stim->AddStimulus(p_init_stim);
				// delete p_init_stim;
			}
			start_time +=(double) rNumStims * (current_cycle_length); // update start time for next set of cycle lengths.
			current_cycle_length-=rCycleLengthReduction; // reduce cycle length.
	    }
	    mpStimulus = p_multi_stim;

   }

    AbstractCardiacCell* CreateCardiacCellForTissueNode(Node<SPACE_DIM>* pNode)
    {
        double x = pNode->GetPoint()[0];

        if (x < 0.03)
        {
        	return new CELLMODEL(this->mpSolver, this->mpStimulus, this->mParameterVector);
        }
        else
        {
            return new CELLMODEL(this->mpSolver, this->mpZeroStimulus, this->mParameterVector);
        }
    }
};

class TestStochasticRapidPacing : public CxxTest::TestSuite
{

public:


    void TestRapidTissueSimulations() throw(Exception)
    {
        // Get command line arguments
        CommandLineArguments* p_args = CommandLineArguments::Instance();
        unsigned argc = *(p_args->p_argc); // has the number of arguments.
        if (PetscTools::AmMaster())
        {
        	std::cout << "# " << argc-1 << " arguments supplied.\n" << std::flush;
        }

        // Define conductivity scale
        double conductivity_scale = 1;
        if (CommandLineArguments::Instance()->OptionExists("--conductivity_scale"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--conductivity_scale");
            conductivity_scale = atof(val);
        }

        // Define spatial discretisation for finite element mesh
        double h = 0.01; // cm
        if (CommandLineArguments::Instance()->OptionExists("--spatial_discretization"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--spatial_discretization");
            h = atof(val);
        }

        // Define ode time step
        double ode_time_step = 0.005; //ms
        if (CommandLineArguments::Instance()->OptionExists("--ode_time_step"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--ode_time_step");
            ode_time_step = atof(val);
        }

        // Define pde time step
        double pde_time_step = 0.01; //ms
        if (CommandLineArguments::Instance()->OptionExists("--pde_time_step"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--pde_time_step");
            pde_time_step = atof(val);
        }

        // Set number of stimuli to apply
        unsigned num_stims = 5;
        if (CommandLineArguments::Instance()->OptionExists("--num_stims"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--num_stims");
            num_stims = atof(val);
        }
        // Check at least one stimulus is applied
        assert(num_stims >= 1);

        // Set cycle length in ms
        unsigned cycle_length = 250;
        if (CommandLineArguments::Instance()->OptionExists("--cycle_length"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--cycle_length");
            cycle_length = atof(val);
        }
        // Check that cycle length > 0
        assert(cycle_length > 0.0);

        // Set cycle length reduction in ms
        unsigned cycle_length_reduction = 10;
        if (CommandLineArguments::Instance()->OptionExists("--cycle_length_reduction"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--cycle_length_reduction");
            cycle_length_reduction = atof(val);
        }
        // Check that cycle length reduction > 0
        assert(cycle_length_reduction > 0.0);

        // Set number of cycle lengths
        unsigned num_cycle_lengths = 2;
        if (CommandLineArguments::Instance()->OptionExists("--num_cycle_lengths"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--num_cycle_lengths");
            num_cycle_lengths = atof(val);
        }
        // Check that num cycle lengths > 0
        assert(num_cycle_lengths > 0u);

        bool stochastics = false;
        if (CommandLineArguments::Instance()->OptionExists("--stochastic"))
        {
        	stochastics = true;
        }

        unsigned cell_model = 0u;
        if (CommandLineArguments::Instance()->OptionExists("--cell_model"))
        {
            char* val = CommandLineArguments::Instance()->GetValueCorrespondingToOption("--cell_model");
            double cell_model_option = atof(val);
            cell_model = boost::lexical_cast<unsigned>(cell_model_option)-1u; // -1u means input is 1-4, code sees 0-3 as per indexing.
        }

        if (cell_model > 3u)
        {
        	if (PetscTools::AmMaster())
        	{
        		std::cout << "Cell model number is " << cell_model <<"\n";
        	}
        	EXCEPTION("Cell model number out of range");
        }

        if (PetscTools::AmMaster())
        {
        	std::cout << "conductivity_scale is "<< conductivity_scale << "\n";
        	std::cout << "spatial_discretization is "<< h << "\n";
        	std::cout << "ode_time_step is "<< ode_time_step << "\n";
        	std::cout << "pde_time_step is "<< pde_time_step << "\n";
        	std::cout << "num_stims is "<< num_stims << "\n";
        	std::cout << "cycle_length is "<< cycle_length << "\n";
        	std::cout << "cell_model is "<< cell_model << "\n";
        }

        DistributedTetrahedralMesh<2,2> mesh;
        unsigned num_elem_x = (unsigned)(1.0/h);  // num elements to make 1cm
        unsigned num_elem_y = (unsigned)(1.0/h);  // num elements to make 1cm
        double stim_mag = -500;
        double stim_dur = 3;
        double area = 0.005;

        mesh.ConstructRectangularMesh(num_elem_x, num_elem_y);
        mesh.Scale(h,h); // Get mesh into units of cm.

        std::string archive_dir_base("StochasticsRapidPacing_archives/archive");
        std::string archive_dir_current;

        // Cell model parameters
		// Guinea Pig apex 1
		std::vector<double> gp_apex_1_params;
		gp_apex_1_params = boost::assign::list_of<double>
		(0.2985)(0.1299)(0.2016)(0.0060)(0.9502)(0.0160)(0.6456)(1.5463)(0.0001)(0.0273)(462.4436)(5.9510)(2.6892)(1.9773)(39.3173)(1.1926)(5.9208)(139.5427)(75.3934)(10.0256)(2.9338)(1.4622)(275.9146)(0.0957)(0.7847)(1.9465)(198.1980)(2.2028)(94.0553);

    	// Guinea Pig apex 2
		std::vector<double> gp_apex_2_params;
		gp_apex_2_params = boost::assign::list_of<double>
		(0.3209)(0.1302)(0.1943)(0.0056)(0.9802)(0.0166)(0.6475)(1.6175)(-0.0008)(0.0410)(416.8280)(6.1000)(3.0565)(2.1917)(30.9108)(0.2569)(6.3650)(150.3936)(76.2596)(10.2019)(2.7270)(1.3532)(276.2747)(0.1094)(0.8426)(2.0059)(203.9794)(2.2932)(103.7665);

		// Guinea Pig base 1
		std::vector<double> gp_base_1_params;
		gp_base_1_params = boost::assign::list_of<double>
		(0.2749)(0.1254)(0.1908)(0.0058)(0.9212)(0.0154)(0.6645)(1.5920)(0.0004)(0.0360)(476.7330)(5.5794)(2.4822)(1.7640)(39.1266)(1.1066)(5.6937)(133.9741)(72.9930)(9.6374)(2.9016)(1.5253)(352.7995)(0.0950)(0.7252)(2.0227)(187.9975)(2.6529)(94.6066);

		// Guinea Pig base 2
		std::vector<double> gp_base_2_params;
		gp_base_2_params = boost::assign::list_of<double>
		(0.3005)(0.1202)(0.1714)(0.0062)(0.9391)(0.0175)(0.7072)(1.5749)(0.0003)(0.0290)(456.3476)(6.0416)(2.8133)(1.9486)(28.8126)(1.2874)(6.1828)(137.4365)(66.5924)(10.0310)(2.7844)(1.3333)(273.7935)(0.0944)(0.7848)(1.9425)(206.9896)(2.8119)(90.4975);

		// Make into 'new' model:
		gp_apex_1_params.push_back(1.1);
		gp_apex_1_params[5] = 0.77;
		gp_apex_1_params[16] = 80;
		gp_apex_1_params[17] = 100;
		gp_apex_1_params[24] = 0.95;

		gp_apex_2_params.push_back(1.1);
		gp_apex_2_params[5] = 0.77;
		gp_apex_2_params[16] = 80;
		gp_apex_2_params[17] = 100;
		gp_apex_2_params[24] = 1.00;

		gp_base_1_params.push_back(1.1);
		gp_base_1_params[5] = 0.77;
		gp_base_1_params[16] = 80;
		gp_base_1_params[17] = 100;
		gp_base_1_params[24] = 0.95;

		gp_base_2_params.push_back(1.1);
		gp_base_2_params[5] = 0.77;
		gp_base_2_params[16] = 80;
		gp_base_2_params[17] = 100;
		gp_base_2_params[24] = 0.95;

		if (stochastics)
		{
			gp_apex_1_params.push_back(0.0012);
			gp_apex_1_params.push_back(0.0012);

			gp_apex_2_params.push_back(0.0015);
			gp_apex_2_params.push_back(0.0020);

			gp_base_1_params.push_back(0.0011);
			gp_base_1_params.push_back(0.0013);

			gp_base_2_params.push_back(0.0011);
			gp_base_2_params.push_back(0.0012);
		}

		std::vector<std::vector<double> > parameter_vectors;
		parameter_vectors.push_back(gp_apex_1_params);
		parameter_vectors.push_back(gp_apex_2_params);
		parameter_vectors.push_back(gp_base_1_params);
		parameter_vectors.push_back(gp_base_2_params);

        // Setup
		HeartConfig::Instance()->SetVisualizeWithMeshalyzer();
        HeartConfig::Instance()->SetSimulationDuration(cycle_length); //ms

        std::stringstream out_dir;
        out_dir << "StochasticsRapid_CL_" << cycle_length << "_cs_" << conductivity_scale << "_cell_model_" << cell_model;

        if (stochastics)
        {
        	out_dir << "_stochastic";
        }

        HeartConfig::Instance()->SetOutputDirectory(out_dir.str());
        HeartConfig::Instance()->SetOutputFilenamePrefix("results");
        HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(ode_time_step,pde_time_step,0.1);
        HeartConfig::Instance()->SetIntracellularConductivities(Create_c_vector(1.4*conductivity_scale*1.171, 1.4*conductivity_scale*1.171));
        HeartConfig::Instance()->SetSurfaceAreaToVolumeRatio(1400.0); // 1/cm
        HeartConfig::Instance()->SetCapacitance(1.0); // uF/cm^2

        std::vector<std::pair<double,double> > apds_requested;
        apds_requested.push_back(std::pair<double, double>(90,0.4)); //repolarisation percentage and threshold
        HeartConfig::Instance()->SetApdMaps(apds_requested);
        std::vector<double> excitation_threshold;
        excitation_threshold.push_back(double (0.4));
        HeartConfig::Instance()->SetUpstrokeTimeMaps(excitation_threshold);
        HeartConfig::Instance()->SetMaxUpstrokeVelocityMaps(excitation_threshold);

		double end_time = 0;
		double pacing_cycle_length = cycle_length;

        for (unsigned stim_counter=0; stim_counter < num_cycle_lengths * num_stims; stim_counter++ )
        {
            // Load problem
            MonodomainProblem<2> *p_monodomain_problem;
            if (stim_counter==0)
            {
            	if (stochastics)
            	{
            		RapidPacingCellFactory<CellWalmsley2015_ParameterVector_Stochastic,2> cell_factory(stim_mag, stim_dur, area, cycle_length, cycle_length_reduction, num_stims, num_cycle_lengths, parameter_vectors[cell_model]);
                	p_monodomain_problem = new MonodomainProblem<2>( &cell_factory );
                    p_monodomain_problem->SetMesh(&mesh);
                    p_monodomain_problem->Initialise();
            	}
            	else
            	{
            		RapidPacingCellFactory<CellWalmsley2015_ParameterVector,2> cell_factory(stim_mag, stim_dur, area, cycle_length, cycle_length_reduction, num_stims, num_cycle_lengths, parameter_vectors[cell_model]);
                	p_monodomain_problem = new MonodomainProblem<2>( &cell_factory );
                    p_monodomain_problem->SetMesh(&mesh);
                    p_monodomain_problem->Initialise();
            	}

            }
            else
            {
                p_monodomain_problem = CardiacSimulationArchiver<MonodomainProblem<2> >::Load(archive_dir_current);
            }

            end_time += pacing_cycle_length;
            std::cout<< "new end_time " << end_time << "\n";

            HeartConfig::Instance()->SetSimulationDuration(end_time); //ms - note this is the time the simulation will end at, not the total duration! Starts at non-zero time

            // now decrease CL for next stim if necessary.
            std::cout << "stim_counter+1 " << stim_counter+1 << "\n";
            std::cout << "num_stims " << num_stims << "\n";
            std::cout << "stim_counter+1 % num_stims " << (stim_counter+1) % num_stims << "\n";

            if (stim_counter > 0 && (stim_counter+1) % num_stims == 0)
            {
            	pacing_cycle_length-=cycle_length_reduction;
            	std::cout << "new pacing cycle length " << pacing_cycle_length << "\n";
            }

            // set new directories to work from
            std::stringstream stringoutput;
            stringoutput << stim_counter << "_CS_" << conductivity_scale;
            std::string stim_counter_string = stringoutput.str();

            archive_dir_current = archive_dir_base + "_" + stim_counter_string;
	    	
			if(stochastics)
	    	{
				archive_dir_current = archive_dir_current + "_stochastic";
	    	}

            HeartConfig::Instance()->SetOutputFilenamePrefix("results_" + stim_counter_string);

            // Solve problem (this does the postprocessing too when HeartConfig options are set).
            p_monodomain_problem->Solve();

            // Save problem to archive
            CardiacSimulationArchiver<MonodomainProblem<2> >::Save(*p_monodomain_problem, archive_dir_current, false);
            std::cout << "Archived to " << archive_dir_current << "\n" << std::flush;

            if (PetscTools::AmMaster()) // Best to have only the master processor messing with files
            {
                // Copy the postprocessing results into the archive folders so they aren't wiped.
                std::vector<std::string> files;
                files.push_back("Apd_90_0pt40_Map");
                files.push_back("MaxUpstrokeVelocityMap_0pt40");
                files.push_back("UpstrokeTimeMap_0pt40");
                for (unsigned i=0; i<files.size(); i++)
                {
                    std::string command = "mv " +  OutputFileHandler::GetChasteTestOutputDirectory() + HeartConfig::Instance()->GetOutputDirectory() + "/output/" + files[i] + ".dat "
                                          + OutputFileHandler::GetChasteTestOutputDirectory() + archive_dir_current + "/" + files[i] + ".dat";

                    // If we are in an PetscTools::AmMaster() block we need to use this instead of EXPECT0.
                    ABORT_IF_NON0(system, command);
                }
            }
            PetscTools::Barrier();

        }// close for loop
	
        // Report time spent on simulations
        HeartEventHandler::Headings();
        HeartEventHandler::Report();
	
    }//close void TestRapidTissueSimulations

};

#endif /* TESTSTOCHASTICSRAPIDPACING_HPP_ */

